import io
import sys
import math
import numpy as np
import pygame
from pygame.locals import *

WINDOW_HEIGHT =  640
WINDOW_WIDTH  = 1280
WINDOW_BOTTOM_MARGIN = 40

BACKGROUND_COLOR = (0,0,0)
TEXT_COLOR = (192,192,192)
MENU_COLOR = (0,192,0)
CITY_NAME_COLOR = (0,128,0)
CITY_POSITION_COLOR = (255, 104, 31)
DISTANCE_COLOR = (230,188,92)
PATH_COLOR = (230,188,92)

LINE_COLORS = [
	(64,64,64),
	(128,128,128),
	(96,96,96), 
	(160, 160, 160) 

]

MENU_TEXT = "F1 View  -  F2 List of Cities  -  F3 Help  -  F4 About -   F10 Quit"
HELP_TEXT = """
Use the following keys :
- F1 to go back to path view
- F2 to see list of cities
- F3 to see help menu
- F4 to see the About window 
- F9 to print path on console
- F10 to quit

To modify the path
- Click on a first city and another one to link them
"""

# ===================================================================
#
#	Class used to handle all data of the TSP problem and draw
#   path and cities in the pygame window
#
# ===================================================================

class TSPVisualizer(object):

	"""
		Constructor
		
		Parameters:
		
			- file_name: name of file from which to read information
			
		Objective:
		
			read number of cities, names of cities, matric of 
			distances between cities, positions of cities and	
	"""
	def __init__(self, file_name):
		self.nbr_cities = 0
		self.list_of_cities = []
		self.list_of_distances = []
		self.list_of_positions = []
		self.list_cities_text = []
		self.path = []
		self.list_of_paths = []
		
		# read data
		if self.read_file( file_name ) == False:
			raise RuntimeError("while reading input file. Check file format")
		
		
		self.list_of_distances = np.array( self.list_of_distances )
		self.list_of_distances = self.list_of_distances.astype(np.float)
		if self.list_of_distances.shape[0] != self.nbr_cities:
			raise RuntimeError("in matrix of distances")
		
		if self.list_of_distances.shape[1] != self.nbr_cities:
			raise RuntimeError("in matrix of distances")
		
		if len(self.path) == 0:
			raise RuntimeError("in path")
		
		if not self.alldiff( self.path ):
			raise RuntimeError("in path, same city appears twice or city does not appear")
		
		if len(self.list_of_positions) == 0:
			self.generate_positions()
		
		# modify data
		for i in range(self.nbr_cities):
			self.path[i] -= 1
			self.list_of_cities[ i ] = str(i+1) + ". " + self.list_of_cities[ i ]
		
		# modify positions of cities to match width and height of window	
		# changement de repère
		vx = []	
		vy = []
		for x,y in self.list_of_positions:
			vx.append(x)
			vy.append(y)
			
		x_min = min(vx)	
		x_max = max(vx)
		y_min = min(vy)	
		y_max = max(vy)
			
		print("x_min=", x_min, ", x_max=", x_max)
		print("y_min=", y_min, ", y_max=", y_max)
		
		XW = WINDOW_WIDTH - 150
		YH = WINDOW_HEIGHT - 100
		
		for i in range(len(self.list_of_positions)):
			x = self.list_of_positions[i][0]
			y = self.list_of_positions[i][1]
		
			x = int( 75 + ( (x - x_min) / (x_max - x_min) * XW) )
			y = int( 50 + ( (y - y_min) / (y_max - y_min) * YH ) )
			self.list_of_positions[i][0] = x
			self.list_of_positions[i][1] = y
			
		self.generate_cities_names_positions()	
			
	"""
	"""		
	def generate_cities_names_positions(self):
		X_center, Y_center = 0, 0
		for c in self.list_of_positions:
			X_center += c[0]
			Y_center += c[1]
		
		X_center = int( X_center / self.nbr_cities )
		Y_center = int( Y_center / self.nbr_cities )
		
		self.X_center = X_center
		self.Y_center = Y_center
		
		print("X_center=", X_center)
		print("Y_center=", Y_center)
		
		i = 0
		for c in self.list_of_positions:
			x, y = c[0], c[1]
			d_x = x - X_center
			d_y = y - Y_center
			distance = math.sqrt( d_x * d_x + d_y * d_y )
			angle = math.atan2(d_y, d_x)
			angle_in_degrees = math.degrees(angle) 
		
			d2 = distance * 1.05
			n_x = X_center + int(d2 * math.cos(angle))
			n_y = Y_center + int(d2 * math.sin(angle))
			n_x, n_y = c[0], c[1]
		
			print("angle = ",  angle, angle_in_degrees)
			
			if 90.0 < angle_in_degrees < 180.0:
				n_x -= 10
			elif 0.0 < angle_in_degrees < 90.0:
				n_x += 10
		
			if 0.0 < angle_in_degrees < 180.0:
				n_y += 20
			else:
				n_y -= 20
				 
			self.list_cities_text.append( (n_x, n_y) )
			i += 1
			
	"""
	
	"""	
	def alldiff(self, l):
		print(l)
		for i in range(1, len(l)+1):
			if not i in l:
				print(" => ", i, " missing ")
				return False
				
		return True
	
	"""
		Read file of data for an instance of the TSP
		
	"""	
	def read_file(self, file_name):
		global WINDOW_WIDTH, WINDOW_HEIGHT
		
		# reading states
		# - 0 : start, read number of cities
		# - 1 : read cities names
		# - 2 : read mmmatrix of distances
		# - 3 : read path to display
		# - 4 : screen size
		# - 5 : read positions of cities (if exist)
		
		state = 0
		
		file = open( file_name )
		for line in file:
			line = line.strip()
			if len(line) == 0:
				continue
			if line[0] == "#":
				continue
			#print(ligne)	
			
			# read number of cities
			if state == 0:
				self.nbr_cities = int( line )
				state = 1
				continue 
				
			# read names of cities	
			if state == 1:
				self.list_of_cities.append( line )
				if len( self.list_of_cities ) == self.nbr_cities:
					state = 2
					continue
					
			# read distances
			if state == 2:
				self.list_of_distances.append( line.split(',' ) )
				if len( self.list_of_distances ) == self.nbr_cities:
					state = 3
					continue

			# read path
			if state == 3:
				array = line.split(',')
				for x in array:
					self.path.append( int(x) )
				state = 4
				continue
				
			# read window size	
			if state == 4:	
				array = line.split(',')
				WINDOW_WIDTH  = int(array[0])
				WINDOW_HEIGHT = int(array[1])
				state = 5
				continue
				
			# read positions of cities
			if state == 5:
				array = line.split(",")
				self.list_of_positions.append( [ float(array[0]),  float(array[1]) ] )
				if len( self.list_of_positions ) == self.nbr_cities:
					state = 5
					continue
				
		file.close()
		
		if state >= 3:
			return True
		else:
			return False
	
	"""
	
	"""		
	def generate_positions(self):
		global WINDOW_WIDTH, WINDOW_HEIGHT
		
		angle_delta = 360 / self.nbr_cities
		angle = 0
		for i in range(self.nbr_cities):
			x = WINDOW_WIDTH // 2 + int( 200 * math.cos( math.radians( angle ) ) ) 
			y = WINDOW_HEIGHT // 2 + int( 200 * math.sin( math.radians( angle ) ) )
			angle += angle_delta
			print("generate ", x, y )
			self.list_of_positions.append( [x,y] )
		
	"""
	
	"""		
	def __str__(self):
		s = str( self.nbr_cities ) + "\n"
		s += "==== names ====\n"
		i = 1
		for city_name in self.list_of_cities:
			s += str(i) + "." + city_name + "\n"
			i += 1
		s += "==== distances ====\n"
		s += np.array2string( self.list_of_distances ) + "\n"
		s += "==== positions ====\n"
		i = 1
		for x,y in self.list_of_positions:
			s += str(i) + ". (" +str(x) +","+str(y) + ")\n"
			i += 1
				
		return s	
		
	"""
		Draw path using cities positions and names
		
	"""	
	def draw(self, window):
		global font_14pt, LINE_COLORS, BURNT_AMBER, SIENNA
		
		distance = 0
		
		# draw lines between cities and compute distance
		for i in range(1, self.nbr_cities):
			src = self.list_of_positions[ self.path[i-1] ]
			dst = self.list_of_positions[ self.path[i] ]
			color = LINE_COLORS[ i % len(LINE_COLORS) ]
			pygame.draw.line(window, color, src, dst, 3)
			distance += self.list_of_distances[ self.path[i-1] ][ self.path[i] ]
		
		src = dst
		dst = self.list_of_positions[ self.path[0] ]
		color = LINE_COLORS[ (i+1) % len(LINE_COLORS) ]
		pygame.draw.line(window, color, src, dst, 3)	
		distance += self.list_of_distances[ self.path[i] ][ self.path[0] ]
		
		# draw cities positions
		for x,y in self.list_of_positions:
			pygame.draw.circle(window, CITY_POSITION_COLOR, (x, y), 5)

		# draw cities names		
		for i in range(0, self.nbr_cities):
			x = self.list_cities_text[ i ][ 0 ]
			y = self.list_cities_text[ i ][ 1 ]
			rendu = font_14pt.render(self.list_of_cities[i], True, CITY_NAME_COLOR)
			rect = rendu.get_rect() 
			rect.center = (x, y)
			pygame.draw.rect(window, BACKGROUND_COLOR, rect)
			window.blit(rendu, rect)
		
		# draw distance	
		distance_str = "Distance " + str(int(distance))
		rendu = font_24pt.render(distance_str, True, DISTANCE_COLOR)
		rect = rendu.get_rect() 
		rect.center = (WINDOW_WIDTH // 2, 20)
		pygame.draw.rect(window, BACKGROUND_COLOR, rect)
		window.blit(rendu, rect)
		
		# draw path
		path_to_display = [ x+1 for x in self.path ]
		rendu = font_14pt.render("Path = " + str(path_to_display), True, PATH_COLOR)
		rect = rendu.get_rect() 
		rect.move_ip(30, WINDOW_HEIGHT-30)
		pygame.draw.rect(window, BACKGROUND_COLOR, rect)
		window.blit(rendu, rect)
		
			


	"""

	"""
	def events(self):
		
		for event in pygame.event.get():
			if event.type == QUIT:
				self.stop_game = True
				
			if event.type == KEYDOWN:
				# Ctrl-C to leave game
				if event.key == pygame.K_c and pygame.key.get_mods() & pygame.KMOD_CTRL:
					self.stop_game = True
				
				if event.key == pygame.K_F10:
					self.stop_game = True
					
				if event.key == K_F1:
					self.window_id = 1
					return None
					
				if event.key == K_F2:
					self.window_id = 2
					return None
					
				if event.key == K_F3:
					self.window_id = 3		
					return None
				
				if event.key == K_ESCAPE:
					self.window_id = 1
					return None
					
				# return key pressed if needed
				return event.key

				
			if event.type == MOUSEBUTTONDOWN:
					return MOUSEBUTTONDOWN
					
		# return (valeur nulle)
		return None   	

	"""
	
	"""
	def modify_current_path(self):
		# record copy of current path into list
		self.list_of_paths.append( self.path.copy() )
		

		index_src_city = self.path.index( self.src_city_id )
		index_dst_city = self.path.index( self.dst_city_id )

		print("============================")
		print("MODIFY PATH")
		print("============================")
		print("Initial path=", self.path)	
		
		print("city src=", self.src_city_id, index_src_city )
		print("city dst=", self.dst_city_id, index_dst_city )
		
		self.path.remove( self.dst_city_id )
		if index_dst_city > index_src_city:
			self.path.insert( index_src_city + 1, self.dst_city_id )	
		else:
			self.path.insert( index_src_city - 1, self.dst_city_id )	
		
		print("Final path=", self.path)	

	"""
	
	"""
	def find_city_from_position(self, xs, ys): 
		i = 0
		for (x,y) in self.list_of_positions:
			if x-10 < xs < x+10:
				if y-10 < ys < y+10:
					return i
			i = i + 1
	
		return -1
	
	
	def modify_path_using_mouse(self, xm, ym, city_id):
		
		if city_id == -1:
			return 
			
		if self.mouse_state == 0:
			self.src_city_id = city_id
			self.mouse_state = 1
			
		elif self.mouse_state == 1:
			self.dst_city_id = city_id
			self.mouse_state = 2
			self.modify_current_path()
			self.mouse_state = 0
	
	
	"""
	
	"""
	def render_path(self):
		pygame.draw.line(window, (127,127,127), (0, WINDOW_HEIGHT), (WINDOW_WIDTH,WINDOW_HEIGHT), 1)
		pygame.draw.line(window, (107,107,107), (0, WINDOW_HEIGHT+1), (WINDOW_WIDTH,WINDOW_HEIGHT+1), 1)
		pygame.draw.line(window, (87,87,87), (0, WINDOW_HEIGHT+2), (WINDOW_WIDTH,WINDOW_HEIGHT+2), 1)
		self.draw_text(MENU_TEXT, 30, WINDOW_HEIGHT+4, MENU_COLOR)
		self.draw( window )
		key = self.events()
		
		if key == K_F9:
			# print path
			print( self.path[0]+1, end='')
			for i in range(1, len(self.path)):
				print( ", " + str(self.path[i]+1), end='')
			print()
			
			
		if key == MOUSEBUTTONDOWN:
			(button_1, button_2, button_3) = pygame.mouse.get_pressed()

			if button_1 == 1:
				mouse = pygame.mouse.get_pos()  
				x_mouse, y_mouse = mouse[0], mouse[1]
				print(x_mouse, y_mouse)
				city_id = self.find_city_from_position( x_mouse, y_mouse ) 
				self.modify_path_using_mouse( x_mouse, y_mouse, city_id )
				
			if button_3 == 1:
				if len(self.list_of_paths) != 0:
					self.path = self.list_of_paths.pop()
				print(self.path)


	"""
	
	"""
	def render_cities(self):
		self.draw_text(MENU_TEXT, 30, WINDOW_HEIGHT+4, MENU_COLOR)
		x = 30
		y = 30
		for i in range(self.nbr_cities):
			rendu = font_14pt.render(self.list_of_cities[i], True, TEXT_COLOR)
			rect = rendu.get_rect() 
			rect.move_ip(x, y)
			window.blit(rendu, rect)
			y += 20
		key = self.events()

	"""
	
	"""
	def render_help(self):
		self.draw_text(MENU_TEXT, 30, WINDOW_HEIGHT+4, MENU_COLOR)
		self.draw_text(HELP_TEXT, 30, 40, TEXT_COLOR)
		key = self.events()	
	


	"""
		
	"""
	def render(self,window):
		self.window = window
		self.stop_game = False
		self.window_id = 1
		
		# used to change path
		self.mouse_state = 0
		self.src_city_id = -1
		self.dst_city_id = -1
						
		while not self.stop_game:
			window.fill( BACKGROUND_COLOR )
			#print(self.window_id, self.stop_game)
				
			if self.window_id == 1:
				self.render_path()
										
			elif self.window_id == 2:
				self.render_cities()
					
			elif self.window_id == 3:
				self.render_help()				

			pygame.display.update()
			pygame.time.delay(5)	


	"""

	"""
	def draw_string(self, s, x, y, color=None):
		global font_24pt
		
		if color == None:
			color = TEXT_COLOR
			
		rendu = font_24pt.render(s, True, color)
		rect = rendu.get_rect() 
		rect.move_ip(x, y)
		pygame.draw.rect(self.window, BACKGROUND_COLOR, rect)
		self.window.blit(rendu, rect)
	
	"""
	
	"""	
	def draw_text(self, s, x, y, color=None):
		global font_24pt

		if color == None:
			color = TEXT_COLOR
		
		array = s.split("\n")
		for s in array:
			rendu = font_24pt.render(s, True, color)
			rect = rendu.get_rect() 
			rect.move_ip(x, y)
			pygame.draw.rect(self.window, BACKGROUND_COLOR, rect)
			self.window.blit(rendu, rect)
			y += 30


		
		
#
# Main program
#
stop_game = False
window_id = 0
file_name = "usa_data.txt"
visu = None

try:
	if len(sys.argv) >= 1:
		file_name = sys.argv[1]
	
	print("@@@@@@@", file_name )
	visu = TSPVisualizer( file_name )
	print(visu)
except RuntimeError as e:
	print("ERROR: {0}".format(e))


window = pygame.display.set_mode( (WINDOW_WIDTH, WINDOW_HEIGHT + WINDOW_BOTTOM_MARGIN) )
pygame.display.set_caption("TSP Visualizer")
pygame.font.init()
#font_24pt = pygame.font.Font("Roboto-Medium.ttf", 24)
#font_14pt = pygame.font.Font("Roboto-Medium.ttf", 14)
font_24pt = pygame.font.Font("PermanentMarker-Regular.ttf", 24)
font_14pt = pygame.font.Font("PermanentMarker-Regular.ttf", 18)

visu.render(window)

pygame.quit()
sys.exit(0)	
